// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//           Jing Deng <jing.deng@colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: siena_types.h,v 1.6 2002/09/30 04:50:25 carzanig Exp $
//
#ifndef SIENA_TYPES_H
#define SIENA_TYPES_H

#include <map>
#include <set>
#include <string>
#include <vector>
#include <iostream>

using namespace std;

/**
 *  The classes defined here represent the Siena data model.  In
 *  essence, this file defines messages with their attributes, and
 *  predicates with their filters and constraints.
 *
 *  <p>These definitions are nice and easy, but need to change in a
 *  somewhat fundamental way.  The idea is to define a message (and
 *  filters, constraints, etc.) as a virtual class, and then to offer
 *  implementations (subclasses).  This would allow us to make the
 *  forwarding table parametric w.r.t. the physical representation of
 *  messages.  The virtual message class should look like this:
 *
 *  <pre>
 *  class sx_message {
 *  public:
 *      class iterator {
 *      public
 *          virtual iterator & operator ++();
 *          virtual sx_attribute & operator *() const;
 *      };
 *
 *      virtual iterator & find_attribute(const char *) const;
 *      virtual iterator & begin() const;
 *      virtual iterator & end() const;
 *  }
 *  </pre>
 *
 *  <p>Work in progress...
 **/

/** operator identifier for <em>Siena</em>
 *
 *  Siena defines some of the most common operators (or binary
 *  relations). <code>sx_operator</code> is the type of operator
 *  identifiers.
 *
 *  Notice that the SFF algorithm supports only SX_eq, SX_lt, and
 *  SX_gt at this time.
 **/
enum sx_operator {
    /** equals 
     **/ 
    SX_eq		= 1, 
    /** less than.
     *
     *  less than. Integers and doubles are ordered as usual, strings
     *  are sorted in lexicographical order. For booleans,
     *  <code>false</code> < <code>true</code>.
     **/
    SX_lt		= 2, 
    /** greater than 
     **/
    SX_gt		= 3, 
    /** greater or equal 
     **/
    SX_ge		= 4, 
    /** less or equal 
     **/
    SX_le		= 5, 
    /** suffix
     *	
     *	suffix. <em>x</em> <code>SX_sf</code> <em>y</em> is true if
     *	<em>y</em> is a suffix of <em>x</em>.  For example,
     *	<code>"software" SX_sf "ware" == true</code>, while
     *	<code>"software" SX_sf "w" == false</code>. <code>SX_sf</code>
     *	is defined for strings only.
     **/
    SX_sf		= 7,
    /** prefix
     *
     *	prefix.  <em>x</em> <code>SX_pf</code> <em>y</em> is true if
     *	<em>y</em> is a prefix of <em>x</em>.  For example,
     *	<code>"software" SX_pf "soft" == true</code>, while
     *	<code>"software" SX_pf "of" ==
     *	false</code>. <code>SX_pf</code> is defined for strings only.
     **/
    SX_pf		= 6,
    /** <em>any</em> (don't care)
     *
     *	<em>any</em> (don't care).  Always <code>true</code>.
     **/
    SX_any		= 8, 
    /** not equal 
     **/
    SX_ne		= 9, 
    /** contains substring
     *
     *	contains substring. <em>x</em> <code>SX_ss</code> <em>y</em>
     *	is true if <em>y</em> is a substring of <em>x</em>.  For
     *	example, <code>"software" SX_ss "war" == true</code>, while
     *	<code>"software" SX_ss "hard" ==
     *	false</code>. <code>SX_ss</code> is defined for strings only.
     **/
    SX_ss		= 10 
};

/** type identifier
 **/
enum sx_type {
    SX_null		= 0,
    SX_string		= 1,
    SX_int		= 2,
    SX_double		= 3,
    SX_bool		= 4
};

//
// I have more optimized representations for these, but for now I will
// use the standard types.
//
typedef string		sx_name;
typedef string		sx_string_t;
typedef int		sx_int_t;
typedef double		sx_double_t;
typedef bool		sx_bool_t;

typedef string		sx_identity;

/** generic Siena value type 
 **/
class sx_value {
private:
    sx_type		_type;
    union {
	sx_string_t *	string_v;
	sx_int_t	int_v;
	sx_bool_t	bool_v;
	sx_double_t	double_v;
    };
public:
    sx_value(): _type(SX_null) {};
    ~sx_value() { if (_type == SX_string) delete(string_v); }
    sx_value(const sx_value & x): _type(x._type) {
	switch (x._type) {
	case SX_string: string_v = new string(*x.string_v); break;
	case SX_int: int_v = x.int_v; break;
	case SX_double: double_v = x.double_v;
	case SX_bool: bool_v = x.bool_v;
	}
    };

    sx_value(sx_int_t x): _type(SX_int) { int_v = x; }
    sx_value(sx_bool_t x): _type(SX_bool) { bool_v = x; }
    sx_value(sx_double_t x): _type(SX_double) { double_v = x; }
    sx_value(const sx_string_t & x): _type(SX_string) { 
	string_v = new string(x); }

    sx_type		type()		const { return _type; };
    const sx_string_t & string_value()	const { return *string_v; };
    sx_int_t		int_value()	const { return int_v; };
    sx_bool_t		bool_value()	const { return bool_v; };
    sx_double_t		double_value()	const { return double_v; };

    sx_value &		operator = (const sx_value & x) {
	switch (x._type) {
	case SX_string: return operator = (*x.string_v);
	case SX_int: return operator = (x.int_v);
	case SX_double: return operator = (x.double_v);
	case SX_bool: return operator = (x.bool_v);
	}
	_type = x._type;
    };

    sx_value &		operator = (const string & s) {
	if (_type != SX_string) {
	    _type =  SX_string;
	    string_v = new string(s);
	} else {
	    *string_v = s;
	}
	return *this;
    }

    sx_value &		operator = (sx_int_t i) {
	if (_type == SX_string) {
	    delete(string_v);
	} 
	_type = SX_int;
	int_v = i;
	return *this;
    }

    sx_value &		operator = (sx_bool_t b) {
	if (_type == SX_string) {
	    delete(string_v);
	} 
	_type = SX_bool;
	bool_v = b;
	return *this;
    }

    sx_value &		operator = (sx_double_t d) {
	if (_type == SX_string) {
	    delete(string_v);
	} 
	_type = SX_double;
	double_v = d;
	return *this;
    }
};

class sx_attribute {
public:
    sx_name	name;
    sx_value	value;
};

class sx_constraint {
public:
    sx_name name;
    sx_operator op;
    sx_value	value;

    sx_constraint() : name(), op(SX_any), value(0) {};

    sx_constraint(const sx_name &n, sx_operator o, const sx_value & v)
	: name(n), op(o), value(v) { };

    sx_constraint(const sx_constraint & x)
	: name(x.name), op(x.op), value(x.value) {} ; 
};

struct constraint_comparator {
    bool operator() (const sx_constraint & x, const sx_constraint & y) {
	return x.name < y.name;
    }
};

class sx_message : public map<sx_name, sx_value> {
};

// 
// a Filter is a *conjunction* ("and") of constraints 
//
// WARNING: the forwarding table ASSUMES that constraints are
// alphabetically ordered, so be careful when you change this
// definition
// 
class sx_filter : public multiset<sx_constraint, constraint_comparator> {
public:
    void add_constraint(const sx_constraint & c) {
	insert(c);
    }
};

// 
// a predicate is a *disjunction* ("or") of filters
//
class sx_predicate : public vector<sx_filter> {
};

extern ostream & operator << (ostream &, const sx_predicate &);
extern ostream & operator << (ostream &, const sx_constraint &);
extern ostream & operator << (ostream &, const sx_message &);
extern ostream & operator << (ostream &, const sx_value &);
extern ostream & operator << (ostream &, const sx_constraint &);

extern bool covers(const sx_predicate &, const sx_message &);
extern bool covers(const sx_filter &, const sx_message &);
extern bool covers(const sx_constraint &, const sx_message &);

#endif
